# Generated by Django 4.2.3 on 2023-12-17 20:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0010_rename_from_ocr_bbox_from_ocr_merged_squashed_0016_alter_ocrrun_result_merged'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ocrboxmodel',
            name='iso1_map',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='ocrmodel',
            name='iso1_map',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='tslmodel',
            name='iso1_map',
            field=models.JSONField(blank=True, null=True),
        ),
    ]
