# Generated by Django 4.2.3 on 2024-02-01 00:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0017_alter_ocrboxmodel_iso1_map_alter_ocrmodel_iso1_map_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='ocrmodel',
            name='processor_name',
            field=models.CharField(max_length=128, null=True),
        ),
        migrations.AddField(
            model_name='ocrmodel',
            name='tokenizer_name',
            field=models.CharField(max_length=128, null=True),
        ),
    ]
