# Generated by Django 4.2.4 on 2023-09-20 11:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0008_ocrboxmodel_entrypoint_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='language',
            name='easyocr',
        ),
        migrations.RemoveField(
            model_name='language',
            name='facebookM2M',
        ),
        migrations.RemoveField(
            model_name='language',
            name='tesseract',
        ),
        migrations.AddField(
            model_name='ocrboxmodel',
            name='iso1_map',
            field=models.JSONField(null=True),
        ),
        migrations.AddField(
            model_name='ocrmodel',
            name='iso1_map',
            field=models.JSONField(null=True),
        ),
        migrations.AddField(
            model_name='tslmodel',
            name='iso1_map',
            field=models.JSONField(null=True),
        ),
    ]
