# Generated by Django 5.1 on 2025-02-10 14:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ocr_translate', '0019_ocrboxmodel_active_ocrmodel_active_tslmodel_active'),
    ]

    operations = [
        migrations.CreateModel(
            name='LoadEvent',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('description', models.CharField(max_length=512)),
                ('date', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.AddField(
            model_name='language',
            name='load_events_dst',
            field=models.ManyToManyField(
                related_name='languages_dst', to='ocr_translate.loadevent'
            ),
        ),
        migrations.AddField(
            model_name='language',
            name='load_events_src',
            field=models.ManyToManyField(
                related_name='languages_src', to='ocr_translate.loadevent'
            ),
        ),
        migrations.AddField(
            model_name='ocrboxmodel',
            name='load_events',
            field=models.ManyToManyField(
                related_name='%(class)s', to='ocr_translate.loadevent'
            ),
        ),
        migrations.AddField(
            model_name='ocrmodel',
            name='load_events',
            field=models.ManyToManyField(
                related_name='%(class)s', to='ocr_translate.loadevent'
            ),
        ),
        migrations.AddField(
            model_name='tslmodel',
            name='load_events',
            field=models.ManyToManyField(
                related_name='%(class)s', to='ocr_translate.loadevent'
            ),
        ),
    ]
